#ifndef ENVGET_C
#define ENVGET_C

#include "envget.h"
#ifdef __linux__
#   include <linux/string.h>	/* for memcmp() */
#else
#   include <string.h>		/* for memcmp() */
#endif
/* NOTE:
   Do not need to save both beg and end of var name within env string.
   Request must be hashed so len is known and can just add as offset.
   Saving both and comparing length first avoids many failed calls to memcmp().
*/
#define u32	unsigned int

struct Mem {
    Byte const * beg;
    Byte const * end;
};
#ifndef __cplusplus
typedef struct Mem Mem;
#endif

#ifdef __alpha__			/* Use this on ev56: 10+4n cycles */
ENV_LKG u32 MemHash(const Mem mem) {
    unsigned long	 h = 0;		/* 64-bit h ==> 4 vs 6 cycles/byte */
    const Byte		*a = (const Byte *) mem.beg;
    while (a < mem.end) {
	h = h << 13 | h >> (64 - 13);
	h ^= (unsigned)(*a++);
    }
    return (h >> 32 ^ h) * 0x415693bbUL;
}
#else					/* Use this on PPro: 10+2.27n cycles */
#define ROLL13(h) h = h << 13 | h >> 19
ENV_LKG u32 MemHash(const Mem mem) {	/* XXX: breaks on length 0 Mems */
    unsigned char const *a = (unsigned char const *) mem.beg;
    unsigned long h = 0;
    unsigned long n = (unsigned char const *) mem.end - a - 1;
    unsigned long d = (n >> 2) + 1;	/* Dividend; +1 ==> unbranched '--' */
    unsigned long m = 3 & n;		/* Modulus */

    switch (m) {
	case 3: do {	ROLL13(h); h ^= (unsigned)(*a++);
	case 2:		ROLL13(h); h ^= (unsigned)(*a++);
	case 1:		ROLL13(h); h ^= (unsigned)(*a++);
	case 0:		ROLL13(h); h ^= (unsigned)(*a++);
		   } while (--d);
    }
    return h * 0x415693bbUL;
}
#endif

#define TABSZ	(sizeof __env / sizeof(Mem))
#define BITS	8			/* 256*2 ptrs = 1/2 page or so */
#define SHF	(32 - BITS)
#define MSK	((1 << BITS) - 1)
#define H1(h)	((h) >> SHF)		/* High order bits */
#define H2(h)	(((h) & MSK) | 1)	/* Low order bits & ODD */

static Mem __env[1<<BITS];

extern Byte **environ;

ENV_LKG Mem *__env_find(const Mem key) {
    u32  h     = MemHash(key),
	 probe = H1(h),
	 incr  = H2(h),
	 len   = key.end - key.beg,
	 last  = probe;
    Mem  *cprobe;
    const Byte *beg;
loop:
    cprobe = __env + probe;
    if (!(beg = cprobe->beg))
	return cprobe;			/* Empty slot => not found */
    if (cprobe->end - beg == len && memcmp(key.beg, beg, len) == 0)
	return cprobe;			/* Matching slot => found */
    probe += incr;			/* Increment probe mod tabsz */
    probe &= TABSZ - 1;
    if (probe == last)
	return (Mem *) 0;		/* All slots scanned => not found */
    goto loop;
}

#if 0
ENV_LKG int __env_clr() {			/* Clear the __env table */
    Mem *cell,
	*end = __env + TABSZ;
    for (cell = __env; cell < end; cell += 4) {
	cell[0].beg = (Byte *) 0;	/* Code only relies on beg == NULL */
	cell[1].beg = (Byte *) 0;
	cell[2].beg = (Byte *) 0;
	cell[3].beg = (Byte *) 0;
    }
    return 0;
}
#endif

int __env_uptodate = 0;			/* Clean vs. dirty */

ENV_LKG int __env_inc(Byte **list) {	/* Incorporate char *[] into table */
    Mem   var, *cell;
    Byte **pv;

    for (pv = list; *pv; pv++) {
	var.beg = *pv;
	var.end = strchr(*pv, '=');
	if (!var.end)			/* Ignore bogus strings */
	    continue;
	cell = __env_find(var);
	if (!cell || cell->beg)		/* Already there. Uh oh! */
	    continue;			/* First entry wins... */
	*cell = var;
    }
    return __env_uptodate = 1;
}

ENV_LKG Byte *envget(Byte const * const _var, Byte const * const _val) {
    Mem        var,
	      *cell;

    if (!__env_uptodate)		/* environ dirty: refresh */
	__env_inc(environ);
    var.beg = _var;
    var.end = strchr(var.beg, '\0');
    cell = __env_find(var);
    return cell && cell->end ?	(Byte *)(cell->end + 1): (Byte *) _val;
}
#endif /* ENVGET_C */
